class PromptConstants:
    SORRY = "Sorry, I'm not able to provide an answer to that question."
    GRAPH_FIELD_SEP = "<SEP>"
    DEFAULT_TUPLE_DELIMITER = "<|>"
    DEFAULT_RECORD_DELIMITER = "##"
    DEFAULT_COMPLETION_DELIMITER = "<|COMPLETE|>"
    DEFAULT_ENTITY_TYPES = ["organization", "person", "geo", "event", "concept"]
